/**
 * Twitter アクセス用クラス
 * @constructor
 */
AccessTwitter = function() {
}

AccessTwitter.prototype = function () {
  
  /**
   * Twitter API URLs
   */
  var URLS = {
    OAUTH: {
      REQUEST_TOKEN: 'https://api.twitter.com/oauth/request_token',
      ACCESS_TOKEN: 'https://api.twitter.com/oauth/access_token',
      AUTHORIZE: 'https://api.twitter.com/oauth/authorize',
      AUTHENTICATE: 'https://api.twitter.com/oauth/authenticate',
    },
    STATUSES: {
      UPDATE: 'https://api.twitter.com/1.1/statuses/update.json',
    },
    SETTINGS: {
      PC: 'https://twitter.com/settings/sessions',
      MOBILE: 'https://mobile.twitter.com/settings/applications',
    }
  };
  
  /**
   * Twitter エラーコード
   * https://developer.twitter.com/en/docs/basics/response-codes.html
   */
  var ERROR_CODES = {
    INVALID_OR_EXPIRED_TOKEN: 89
  };
  
  /**
   * アクセストークン保存(Local Storage)のためのキー
   * @type {String}
   */
  var KEY_ACCESS_KEY = 'twitter_access_key';
  
  /**
   * OAuth オブジェクトのインスタンス
   * @type {OAuth}
   */
  var oauth = null;
  
  /**
   * Twitter API にアクセスする際のオプション情報
   * @type {Object.<string, string>}
   */
  var options;
  
  // ===== 以下、Private 関数群 =====
  /**
   * アクセスキーを保存
   */
  function saveAccessKey(obj) {
    if (obj == null) {
      localStorage.removeItem(KEY_ACCESS_KEY);
    }
    else if (typeof obj === 'object') {
      localStorage.setItem(KEY_ACCESS_KEY, JSON.stringify(obj));
    }
  }
  
  /**
   * アクセスキーの取得
   */
  function loadAccessKey() {
    return JSON.parse(localStorage.getItem(KEY_ACCESS_KEY));
  }
  
  /**
   * OAuthのインスタンス取得
   */
  function getOAuth() {
    if (!oauth) {
      oauth = OAuth(options);
      
      var accessKey = loadAccessKey();
      debugLog('stored access key');
      debugLog(accessKey);
      
      if (accessKey) {
        oauth.setAccessToken(accessKey.accessTokenKey, accessKey.accessTokenSecret);
      }
    }
    return oauth;
  }
  
  /**
   * OAuthのインスタンス設定
   */
  function setOAuth(obj) {
    oauth = obj;
  }
  
  /**
   * jsOAuthのレスポンスからTwitterのエラーコードを取得
   */
  function getTwitterErrorCode(oauthResponse) {
    var errors = JSON.parse(oauthResponse.text).errors;
    var code = (errors instanceof Array && errors.length > 0) ? errors[0].code : 0;
    return (typeof code === 'number') ? code : -1;
  }
  
  /**
   * アクセストークン確認
   */
  function hasAccessTokenInner() {
    var key = getOAuth().getAccessTokenKey();
    var secret = getOAuth().getAccessTokenSecret();
    return isNotEmpty(key) && isNotEmpty(secret);
  }
  
  /**
   * アクセストークンの破棄
   */
  function clearAccessTokenInner() {
    debugLog('clear stored access key');
    saveAccessKey(null);
    setOAuth(null);
  }
  
  /**
   * ログイン内部処理
   */
  function loginInner(isForce, callback) {
    debugLog('login start...');
    
    if (typeof callback !== 'function') {
      callback = function(){};
    }
    
    // Twitter API: リクエストトークン取得
    getOAuth().get(URLS.OAUTH.REQUEST_TOKEN,
      function (res) {
        debugLog('success get request token.');
        debugLog(JSON.stringify(res));
        
        gotRequestToken(res, isForce, callback);
      }, 
      function(res) {
        debugLog('failed get request token.');
        debugLog(JSON.stringify(res));
        
        if (getTwitterErrorCode(res) === ERROR_CODES.INVALID_OR_EXPIRED_TOKEN) {
          debugLog('Access token was expired.');
          
          // 破棄
          clearAccessTokenInner();
          
          // ログイン
          loginInner(isForce, callback);
        }
        else {
          callback(false);
        }
      }
    );
  }
  
  /**
   * ログアウト内部処理
   */
  function logoutInner() {
    debugLog('logout start...');
    
    var url = (options.mobile === true) ? URLS.SETTINGS.MOBILE : URLS.SETTINGS.PC;
    var wnd = window.open(url, '_blank', 'location=yes');
    wnd.addEventListener('loaderror', function (evt) {
        debugLog('window loaderror');
        debugLog(evt);
    });
    wnd.addEventListener('loadstart', function(evt) {
        debugLog('window loadstart');
        debugLog(evt);
        
        if (evt.url.indexOf('?logout') !== -1) {
          // ログアウトされたので、トークンを削除する
          clearAccessTokenInner();
          wnd.close();
        }
    });
  }
  
  /**
   * ツイート内部処理
   */
  function tweetInner(tweetText, location, callback) {
    debugLog('tweet start...');
    
    if (typeof callback !== 'function') {
      callback = function(){};
    }
    
    var param = {
      'status': tweetText,
      'trim_user': 'true'
    };
    
    if (location && location.latitude && location.longitude) {
      param.lat = location.latitude;
      param['long'] = location.longitude;
    }
    
    // Twitter API: ステータス更新
    getOAuth().post(URLS.STATUSES.UPDATE, param,
      function (res) {
        debugLog('success status update.');
        debugLog(JSON.stringify(res));
        
        callback(true);
      },
      function (res) {
        debugLog('failed status update.');
        debugLog(JSON.stringify(res));
        
        if (getTwitterErrorCode(res) === ERROR_CODES.INVALID_OR_EXPIRED_TOKEN) {
          debugLog('Access token was expired.');
          
          // 破棄
          clearAccessTokenInner();
          
          // 要ログイン
          loginInner(true, function(result) {
              if (result) {
                tweetInner(tweetText, location, callback);
              }
              else {
                callback(false);
              }
          });
        }
        else {
          callback(false);
        }
    });
  }
  
  /**
   * リクエストトークン取得後の処理
   */
  function gotRequestToken(res, isForce, callback) {
    debugLog('Getting authorization...' + res.text);
    
    // Twitter API: アプリ認証
    var requestParams = res.text;
    var wnd;
    var url = (isForce ? URLS.OAUTH.AUTHORIZE : URLS.OAUTH.AUTHENTICATE) + '?';
    url += requestParams + (isForce ? '&force_login=true' : '');
    
    wnd = window.open(url, '_blank', 'location=yes');
    wnd.addEventListener('loaderror', function (evt) {
        debugLog('window loaderror: ' + evt.url);
        
        var tmpurl='http://localhost/';
        wnd.close();
        wnd=window.open(tmpurl, '_blank', 'location=yes');
        
        var params = getURLParams(evt.url);
        var verifier = params['oauth_verifier'];
        if (verifier == null) {
          verifier = '';
        }
        
        // Twitter API: アクセストークン取得
        getOAuth().get(URLS.OAUTH.ACCESS_TOKEN + '?oauth_verifier=' + verifier + '&' + requestParams,
          function (res) {
            debugLog('success get access token.');
            debugLog(JSON.stringify(res));
            
            gotAccessToken(res, callback, wnd);
          }, 
          function(res) {
            debugLog('failed get access token.');
            debugLog(JSON.stringify(res));
            
            wnd.close();
            callback(false);
          }
        );
    });
    
    wnd.addEventListener('loadstart', function (evt) {
        debugLog('window loadstart');
        debugLog(evt);
        
        if (evt.url.indexOf('/intent/session') !== -1) {
          // ログアウトされたので、トークンを削除する
          clearAccessTokenInner();
          return;
        }
        else if (evt.url.indexOf(options.callbackUrl) !== 0) {
          // options.callbackUrl から始まる URL 以外の場合はその画面を表示するので処理しない
          return;
        }
        
        if (!wnd) {
          // wnd が null なので処理中断
          return;
        }
        
        var params = getURLParams(evt.url);
        var verifier = params['oauth_verifier'];
        if (verifier == null) {
          verifier = '';
        }
        
        // Twitter API: アクセストークン取得
        getOAuth().get(URLS.OAUTH.ACCESS_TOKEN + '?oauth_verifier=' + verifier + '&' + requestParams,
          function (res) {
            debugLog('success get access token.');
            debugLog(JSON.stringify(res));
            
            gotAccessToken(res, callback, wnd);
          }, 
          function(res) {
            debugLog('failed get access token.');
            debugLog(JSON.stringify(res));
            
            wnd.close();
            callback(false);
          }
        );
    });
  }
  
  /**
   * アクセストークン取得後の処理
   */
  function gotAccessToken(res, callback, wnd) {
    debugLog('Getting AccessToken... ' + res.text);
    
    var accessParams = getURLParams(res.text);
    getOAuth().setAccessToken(accessParams.oauth_token, accessParams.oauth_token_secret);
    
    saveAccessKey({
        accessTokenKey: accessParams.oauth_token,
        accessTokenSecret: accessParams.oauth_token_secret
    });
    
    wnd.close();
    callback(true);
  }
  
  /**
   * 空文字列チェック
   */
  function isEmpty(str) {
    return (typeof str !== 'string') || (str.length === 0);
  }
  function isNotEmpty(str) {
    return !isEmpty(str);
  }
  
  /**
   * URLクエリパラメータを分解して取得
   */
  function getURLParams(url) {
    var params = {};
    if (typeof url !== "string" || url.length === 0) {
      return params;
    }
    
    var index = url.indexOf('?');
    index = (index >= 0) ? index + 1 : 0;
    
    var queries = url.substring(index).split('&');
    for (var i = 0; i < queries.length; i++) {
      var kv = queries[i].split('=');
      params[kv[0]] = kv[1];
    }
    
    return params;
  }
  
  /**
   * デバッグログ出力
   */
  function debugLog(msg) {
    if (options.debug === true && console && typeof console.log === 'function') {
      console.log(msg);
    }
  }
  
  // ===== 以下、Public 関数群 =====
  return {
    /**
     * 初期化
     * @param {Object.<string, Object>} config 設定
     */
    init: function(config) {
      // Twitter OAuth接続のため、optionsには以下のキー/値を設定してください
      // {
      //   consumerKey: '<Your Twitter Consumer Key (API key)>',
      //   consumerSecret: '<Your Twitter Consumer Secret (API Secret)>',
      //   callbackUrl: '<Your Twitter Callback URL>'
      // }
      options = (typeof config === 'object') ? config : {};
    },
    
    /**
     * アクセストークンの所有確認
     * @return {Boolean} アクセストークンを保持していればtrue, それ以外ならfalse
     */
    hasAccessToken: function() {
      return hasAccessTokenInner();
    },
    
    /**
     * アクセストークンの破棄
     */
    clearAccessToken: function() {
      return clearAccessTokenInner();
    },
    
    /**
     * Twitter ログイン処理
     * @param {Function} callback ログイン処理完了後に実行するコールバック関数
     */
    login: function (callback) {
      loginInner(false, callback);
    },
    
    /**
     * Twitter ログイン処理
     * ログイン済みでも強制で再認証させる
     * @param {Function} callback ログイン処理完了後に実行するコールバック関数
     */
    loginForce: function (callback) {
      loginInner(true, callback);
    },
    
    /**
     * Twitter ログアウト処理
     */
    logout: function() {
      logoutInner();
    },
    
    /**
     * 指定の内容をツイートする
     * @param {string} tweetText 投稿内容
     * @param {Object.<string, float>} location 現在地情報（緯度/経度）
     * @param {Function} callback ツイート処理完了後に実行するコールバック関数
     */
    tweet: function (tweetText, location, callback) {
      if (!hasAccessTokenInner()) {
        loginInner(false, function(result) {
            if (result) {
              tweetInner(tweetText, location, callback);
            }
            else {
              callback(false);
            }
        });
      }
      else {
        tweetInner(tweetText, location, callback);
      }
    }
  };
}();
